#include "i3d/image3d.h"
#include "filter.h"

using namespace i3d;

void Filter_Xaxis(i3d::Image3d<i3d::GRAY8> const &in,
						i3d::Image3d<float> &out,
						const float filter[3])
{
	//image is too small for this simple filtering?
	//(and for any reasonable filtering as well...)
	if (in.GetSizeX() < 3) return;

	//initiate output image (allocates memory, sets resolution etc.)
	out.CopyMetaData(in);

	//working pointers
	const GRAY8 *i=in.GetFirstVoxelAddr();
	float *o=out.GetFirstVoxelAddr();

	for (size_t y=0; y < in.GetSizeY(); ++y) {
		//the first pixel on xLine, has to treat boundary situation
		*o=(float)*i * filter[1] + (float)*(i+1) * filter[2];
		++o; ++i;

		//the middle pixels, far from boundaries...
		for (size_t x=2; x < in.GetSizeX(); ++x) {
			*o=  (float)*(i-1) * filter[0]
				+ (float)*i * filter[1]
				+ (float)*(i+1) * filter[2];
			++o; ++i;
		}

		//the last pixel on xLine, has to treat boundary situation
		*o=(float)*(i-1) * filter[0] + (float)*i * filter[1];
		++o; ++i;
	}
}

void Filter_Xaxis(i3d::Image3d<float> const &in,
						i3d::Image3d<float> &out,
						const float filter[3])
{
	//image is too small for this simple filtering?
	//(and for any reasonable filtering as well...)
	if (in.GetSizeX() < 3) return;

	//initiate output image (allocates memory, sets resolution etc.)
	out.CopyMetaData(in);

	//working pointers
	const float *i=in.GetFirstVoxelAddr();
	float *o=out.GetFirstVoxelAddr();

	for (size_t y=0; y < in.GetSizeY(); ++y) {
		//the first pixel on xLine, has to treat boundary situation
		*o=*i * filter[1] + *(i+1) * filter[2];
		++o; ++i;

		//the middle pixels, far from boundaries...
		for (size_t x=2; x < in.GetSizeX(); ++x) {
			*o=  *(i-1) * filter[0]
				+ *i * filter[1]
				+ *(i+1) * filter[2];
			++o; ++i;
		}

		//the last pixel on xLine, has to treat boundary situation
		*o=*(i-1) * filter[0] + *i * filter[1];
		++o; ++i;
	}
}


void Filter_Yaxis(i3d::Image3d<i3d::GRAY8> const &in,
						i3d::Image3d<float> &out,
						const float filter[3])
{
	//image is too small for this simple filtering?
	//(and for any reasonable filtering as well...)
	if (in.GetSizeY() < 3) return;

	//time saver ;-)
	const size_t xLine=in.GetSizeX();

	//initiate output image (allocates memory, sets resolution etc.)
	out.CopyMetaData(in);

	//working pointers
	const GRAY8 *i=in.GetFirstVoxelAddr();
	float *o=out.GetFirstVoxelAddr();

	//the first pixel in the y-axis (column), has to treat boundary situation
	for (size_t x=0; x < xLine; ++x) {
		*o=(float)*i * filter[1] + (float)*(i+xLine) * filter[2];
		++o; ++i;
	}

	//the middle pixels, far from boundaries...
	for (size_t y=2; y < in.GetSizeY(); ++y) {
		for (size_t x=0; x < xLine; ++x) {
			*o=  (float)*(i-xLine) * filter[0]
				+ (float)*i * filter[1]
				+ (float)*(i+xLine) * filter[2];
			++o; ++i;
		}
	}

	//the last pixel in the y-axis (column), has to treat boundary situation
	for (size_t x=0; x < xLine; ++x) {
		*o=(float)*(i-xLine) * filter[0] + (float)*i * filter[1];
		++o; ++i;
	}
}


void Filter_Yaxis_slower(i3d::Image3d<i3d::GRAY8> const &in,
						i3d::Image3d<float> &out,
						const float filter[3])
{
	//image is too small for this simple filtering?
	//(and for any reasonable filtering as well...)
	if (in.GetSizeY() < 3) return;

	//time saver ;-)
	const size_t xLine=in.GetSizeX();

	//initiate output image (allocates memory, sets resolution etc.)
	out.CopyMetaData(in);

	for (size_t x=0; x < xLine; ++x) {
		//working pointers
		const GRAY8 *i=in.GetVoxelAddr(x,0,0);
		float *o=out.GetVoxelAddr(x,0,0);

		//the first pixel in the y-axis (column), has to treat boundary situation
		*o=(float)*i * filter[1] + (float)*(i+xLine) * filter[2];
		o+=xLine; i+=xLine;

		//the middle pixels, far from boundaries...
		for (size_t y=2; y < in.GetSizeY(); ++y) {
			*o=  (float)*(i-xLine) * filter[0]
				+ (float)*i * filter[1]
				+ (float)*(i+xLine) * filter[2];
			o+=xLine; i+=xLine;
		}

		//the last pixel in the y-axis (column), has to treat boundary situation
		*o=(float)*(i-xLine) * filter[0] + (float)*i * filter[1];
	}
}


void UnsharpMask(i3d::Image3d<i3d::GRAY8> const &in,
					  i3d::Image3d<i3d::GRAY8> &out)
{
	//temporary images for the 1D Laplacians (2nd derivatives)
	i3d::Image3d<float> xFiltered,yFiltered;
	float filter[3]={-1.f,2.f,-1.f};

	//the helper filterings
	Filter_Xaxis(in,xFiltered,filter);
	Filter_Yaxis(in,yFiltered,filter);

	//prepare output image
	out.CopyMetaData(in);

	//note: all images in, xFiltered, yFiltered, and out
	//are of the same sizes

	//pointers...
	const GRAY8 *i=in.GetFirstVoxelAddr();
	float *x=xFiltered.GetFirstVoxelAddr();
	float *y=yFiltered.GetFirstVoxelAddr();
	GRAY8 *o=out.GetFirstVoxelAddr();

	//the unsharp mask
	for (size_t c=0; c < in.GetImageSize(); ++c, ++i, ++x, ++y, ++o) {
		float tmp=(float)*i + 0.5f * (*x + *y);
		if (tmp < 0.) tmp=0.;
		else if (tmp > 255.) tmp=255.;
		*o=(GRAY8)tmp;
	}
}

void UnsharpMask2(i3d::Image3d<i3d::GRAY8> const &in,
					   i3d::Image3d<i3d::GRAY8> &out)
{
	//temporary images for the 1D Laplacians (2nd derivatives)
	i3d::Image3d<float> xFiltered,yFiltered;
	float filter[3]={-1.f,2.f,-1.f};

	//the helper filterings
	Filter_Yaxis(in,yFiltered,filter);
	Filter_Xaxis(yFiltered,xFiltered,filter);

	//prepare output image
	out.CopyMetaData(in);

	//note: all images in, xFiltered, yFiltered, and out
	//are of the same sizes

	//pointers...
	const GRAY8 *i=in.GetFirstVoxelAddr();
	float *x=xFiltered.GetFirstVoxelAddr();
	GRAY8 *o=out.GetFirstVoxelAddr();

	//xFiltered.SaveImage("images/filtr-X.ics");

	//the unsharp mask
	for (size_t c=0; c < in.GetImageSize(); ++c, ++i, ++x, ++o) {
		float tmp=(float)*i + 0.7f * (*x);
		if (tmp < 0.) tmp=0.;
		else if (tmp > 255.) tmp=255.;
		*o=(GRAY8)tmp;
	}
}
