#include <iostream>
#include <time.h>

#include "i3d/image3d.h"
#include "filter.h"

using namespace std;
using namespace i3d;


template <typename VOXEL>
int CompareImages(Image3d<VOXEL> const &A,
						Image3d<VOXEL> const &B)
{
	//porovnani, zda jsou oba obrazy obsahove stejne
	const VOXEL* a=A.GetFirstVoxelAddr();
	const VOXEL* aS=a+A.GetImageSize();
	const VOXEL* b=B.GetFirstVoxelAddr();
	double sum=0;
	while (a != aS) {
		sum+=abs( (double)*a - (double)*b );
		++a; ++b;
	}
	cout << "suma abs. rozdilu obrazu je " << sum << endl;

	return((int)sum);
}


/**
z Wikipedie:

The model used for HDTV developed by the ATSC uses different color
coefficients, computing the luma component as:

    Y' = 0.2126 R + 0.7152 G + 0.0722 B

http://en.wikipedia.org/wiki/Grayscale#Converting_color_to_grayscale
*/
void RGBtoGRAY8(const Image3d<RGB>& in,Image3d<GRAY8>& out) {
	out.CopyMetaData(in);

	const RGB *rgb=in.GetFirstVoxelAddr();
	GRAY8 *g=out.GetFirstVoxelAddr();

	for (size_t i=0; i < in.GetImageSize(); ++i) {
		float tmp=0.2126f * (float)rgb->red
			+ 0.7152f * (float)rgb->green
			+ 0.0722f * (float)rgb->blue;
		*g=(GRAY8)tmp;

		++g; ++rgb;
	}

}


int main(void) {
	try {
		//chytani vyjimek, umoznuje zobrazit chybova hlaseni knihoven i3dlibs
		//tato konstrukce rika, ze v tomto bloku {...} se maji chytat vyjimky

		#define IMGFILE	"images/castle2"
		Image3d<RGB> img;
		img.ReadImageBMP(IMGFILE".bmp");
		Image3d<GRAY8> in;

		//prevede na odstiny sede a vymaze ten barevny z pameti
		RGBtoGRAY8(img,in);
		img.DisposeData();
		in.SaveImageBMP(IMGFILE"_BW.bmp");
		
		Image3d<float> outA,outB;
		float filter[3]={1.f,-2.f,1.f};

		cout << "zpracovavam obraz o rozmerech: " << in.GetSize() 
			<< " a velikosti " << in.GetImageSize()*sizeof(GRAY8)/(1024*1024) << "MB\n";

		clock_t t1,t2;
		//testovani jednotlivych filtru
		Image3d<GRAY8> tmp;

		t1=clock();
		Filter_Xaxis(in,outA,filter);
		t2=clock();

		cout << "doba vypoctu pro osu X byla: "
			<< (float)(t2-t1)/(float)CLOCKS_PER_SEC << " sekund\n";

		FloatToGray(outA,tmp);
		tmp.SaveImageBMP(IMGFILE"_filtr-X.bmp");


		t1=clock();
		Filter_Yaxis(in,outA,filter);
		t2=clock();

		cout << "doba vypoctu pro osu Y byla: "
			<< (float)(t2-t1)/(float)CLOCKS_PER_SEC << " sekund\n";

		FloatToGray(outA,tmp);
		tmp.SaveImageBMP(IMGFILE"_filtr-Y.bmp");

		t1=clock();
		Filter_Yaxis_slower(in,outB,filter);
		t2=clock();

		cout << "doba vypoctu pro osu Y byla: "
			<< (float)(t2-t1)/(float)CLOCKS_PER_SEC << " sekund\n";

		FloatToGray(outB,tmp);
		tmp.SaveImageBMP(IMGFILE"_filtr-Y_slower.bmp");

		//melo by byt 0
		CompareImages(outA,outB);


		//test unsharpmaskovani:
		Image3d<GRAY8> out;
		t1=clock();
		UnsharpMask2(in,out);
		t2=clock();

		cout << "doba vypoctu UnsharpMask byla: "
			<< (float)(t2-t1)/(float)CLOCKS_PER_SEC << " sekund\n";

		out.SaveImageBMP(IMGFILE"_BW_unsharpmask.bmp");

		//nemelo by byt nula
		CompareImages(in,out);


		//pokud se program dostal az sem, nenastal zadny problem
		return(0);

	} //konec bloku "try" - chytani ruznych vyjimek
	catch (const char *e)
	{
		cout << e << endl;
	}
	catch (std::string &e)
	{
		cout << e << endl;
	}
	catch (i3d::IOException& e)
	{
		cout << e << endl;
	}
	catch (i3d::InternalException& e)
	{
		cout << e << endl;
	}
	catch (std::bad_alloc&)
	{
		cout << "Not enough memory." << endl;
	}
	catch (...)
	{
		cout << "System exception." << endl;
	}

	//pokud se program dostal sem, nastal predtim nejaky problem
	return(1);
}
